;+ 
; NAME:
; mike_slitflat
;     Version 1.1
;
; PURPOSE:
;    Stores slit profile and gradient along each order from twilight flats.
;    This routine is critical for performing ideal sky subtraction,
;    especially given the short slit length.  The following steps are
;    performed in mike_slitflat_work:
;    
;    1.  Fit and subtract the scattered light in the twilight flat
;    2.  Extract a boxcar profile down the center of the flat
;    3.  Loop on orders
;    4.  Calculate the slit angle at each pixel in the order
;    5.  Calculate the Jacobian 
;    6.  Fit a bspline to the profile
;    7.  Run diagnsotics on the fit
;    8.  Save the good ones to profile0 and profile1 tags
;
; CALLING SEQUENCE:
;   
;  mike_slitflat, mike, setup, side, [/chk, /clobber]
;
; INPUTS:
;   mike     -  MIKE structure
;   setup    -  Setup identifier 
;   [side]   -  Blue (1), Red (2), or both [1,2L]    (Default: [1,2L])
;
; RETURNS:
;
; OUTPUTS:
;   Fills in the profile0 and profile1 tags in the Order structure
;
; OPTIONAL KEYWORDS:
;  /CHK  - Show the profiles and fits order by order
;  /CLOBBER - Clobber an previous work
;  TFLAT_FIL - Name for TFLAT file
;  JACOB_FIL  - Output name for Jacobian matrix
;  PROFILE_FIL - Output name for profile fits
;  DETILT    - Remove a linear tilt from the Cross-section fit
;
;  NXBKT  -  Number of x breakpoints for scattered light fit 
;  NYBKT  -  Number of x breakpoints for scattered light fit 
;
; OPTIONAL OUTPUTS:
;
; COMMENTS:
;
; EXAMPLES:
;   mike_slitflat, mike, 1
;
; PROCEDURES/FUNCTIONS CALLED:
;  mike_ordermask
;  mike_fitgap
;  mike_qw
;  mike_slitflat_work
;
; REVISION HISTORY:
;   ??-??-2003 Written by SB
;-
;------------------------------------------------------------------------------

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
function mike_slitflat_work, tflat, tflativar, ordr_str, $
         maskimage=maskimage, tflat_file=tflat_file, gapfit=gapfit, $
         chk=chk, jacobian=jacobian, residual=residual, prof_image=prof_image, $
         nxbkpt=nxbkpt, nybkpt=nybkpt, qa_str=qa_str, detilt=detilt

     if NOT keyword_set(slit_samp) then slit_samp = 0.09

     if keyword_set(tflat_file) then begin        
       tflat     = xmrdfits(tflat_file,0)
       tflativar = xmrdfits(tflat_file,1)
     endif

;    Colors
     clr = getcolor(/load)

     sz        = size(tflat, /dimen)
     n_orders = n_elements(ordr_str)

;
;    Mapping out orders and gaps, this is kind of slow
;
     print, 'Calling mike_ordermask...', format='(a,$)'
     maskimage = mike_ordermask(sz[0], sz[1], ordr_str, trim=0.5)
     print, 'Done.'

;
;    Fit scattered light in order gaps
;

     if NOT keyword_set(nxbkpt) then nxbkpt=5
     if NOT keyword_set(nybkpt) then nybkpt=5
     print, 'Fitting Scattered Light...', nxbkpt, nybkpt, format='(a,i4,i4,$)'

     t0 = systime(1)
     gapfit = tflat* 0.
     x = findgen(sz[0])
     for i=0,sz[1]-1 do begin & $
       inorder = where(maskimage[*,i] LT 0 AND tflativar[*,i] GT 0, nin) & $
       if nin GT 70 then begin & $
          temp_row_set = bspline_iterfit(inorder, tflat[inorder,i], $
               everyn=nin/nxbkpt, invvar=tflativar[inorder,i], /silent, $
               maxrej=1, upper=10, lower=10) & $
          gapfit[*,i] = bspline_valu((x > inorder[0]) < inorder[nin-1], $
                                     temp_row_set) & $
       endif & $
     endfor


;     gapfit = mike_fitgap(tflat, tflativar, maskimage, $
;         nxbkpt=nxbkpt, nybkpt=nybkpt)
     print, '  Done. ', systime(1)-t0, $
            min(gapfit), max(gapfit), format='(a,f9.2, f9.2,f9.2)'

     tflat_sub = tflat - gapfit
;     trans_tflat = transpose(tflat - gapfit)
;     trans_ivar = transpose(tflativar)

;
;    Whats up with this jacobian?
;
;       |                  |
;       |  dx/ds    dy/ds  |
;       |                  |
;       |  dx/dl    dy/dl  |
;       |                  |
;
;

   chebypoly = n_elements(ordr_str[0].lhc)
   ncol= (size(maskimage))[1]
   nrow= n_elements(ordr_str[0].lhedg)
   nord = n_elements(ordr_str)
   jacobian = tflat * 0.0
   residual = tflat * 0.0
   prof_image = tflat * 0.0

   tflat_spec = extract_arc(tflat_sub, ordr_str)
;   new_struct = struct_addtags(ordr_str, $
;         replicate({ profile0: fltarr(251), profile1: fltarr(251)},nord))

   profile_x        = findgen(251)/100. - 1.25
   ordr_str.profile0 = 0
   ordr_str.profile1 = 0
   qa_str = 0

   if keyword_set(chk) then $
   print, ' i Ord   N_ft   N_1d   N_2d  linear_coeffs  ----chi_1d----   ----chi_2d----'

 
   ;; Loop on the Orders
   for q=0,nord-1 do begin

      inorder = where(maskimage EQ ordr_str[q].order AND tflativar GT 0, nin)
      print, q, ordr_str[q].order, format='(i2, i4, $)'
      if nin LE nrow then begin
          print, '...not enough pixels in order'
          continue
      endif 

      ystart = 1.0d*(inorder / ncol)
      ordrcen =  (ordr_str[q].lhedg[ystart] + ordr_str[q].rhedg[ystart])/2.0
      xstart = 1.0d*(inorder mod ncol)  - ordrcen
      ycol = dindgen(sz[1])
      slope_spline = spl_init(ycol, double(ordr_str[q].arc_m))

      ;; Calculate wavelengths (and the slope at each x,y pair)
      ywave = mike_qw(xstart, ystart, ordr_str[q].arc_m, arc_slope=arc_slope)


      y = ((2*ystart - nrow) / nrow)
      slit_length = ordr_str[q].rhedg[ystart] - ordr_str[q].lhedg[ystart]
      slit_length[*] = ordr_str[q].rhedg[nrow/2] - ordr_str[q].lhedg[nrow/2]
      slit_frac   = 2.0* xstart / slit_length
      test_deriv = fchebyshev_deriv1(y,chebypoly, /ideriv) 
      test_coeff = ((ordr_str[q].lhc + ordr_str[q].rhc) + $
                     (slit_frac * (ordr_str[q].rhc-ordr_str[q].lhc)))/ nrow
      trace_deriv = test_deriv # test_coeff

      ;; Jacobian
      dydl = 1.0/sqrt(1.0 + trace_deriv^2)
      dxdl = trace_deriv * dydl
      dxds = 1.0/sqrt(1.0 + arc_slope^2)
      dyds = arc_slope* dxds

      jacobian_ord = dxds * dydl - dxdl * dyds
      jacobian[inorder] = jacobian_ord

      ;;
      ;;  now estimate flux_tflat(lambda)
      ;;     
      N = n_elements(inorder)
    
      xsort = sort(slit_frac)
;     tflat_norm = tflat_sub[inorder] / jacobian_ord
      tflat_norm = tflat_sub[inorder] 
      
;
;    good will NOT work on dual slit... we need to find a way to get
;    pixel
;
      xedge = xstart + ordrcen
      good = where(abs(slit_frac) LT 0.7 AND $
                   xedge GT 4 AND xedge LT ncol-4 AND $
                   tflativar[inorder] GT 0,ngood)

       maxy = max(ywave[good])
       miny = min(ywave[good]) 
      everyn = long(0.8*ngood/(maxy - miny))
      ys = sort(ywave[good])
      bset = bspline_iterfit(ywave[good[ys]], tflat_norm[good[ys]], $
                      everyn=everyn, /groupbadpix, upper=10, lower=10, $
                      invvar=tflativar[inorder[good[ys]]], /silent)
      tflat_fit = bspline_valu(((ywave < maxy) > miny), bset)
;      linterp, ycol, tflat_spec[*,q], ywave, tflat_fit

;
;  Now it's time for the slit profile
;

      profile = tflat_norm / (tflat_fit + (tflat_fit EQ 0)) * (tflat_fit GT 20)
;     profile_ivar = tflativar * jacobian_ord^2 * tflat_fit^2 * $
;                    (profile GT -0.1 AND profile LT 1.5 AND tflat_fit GT 0)
      profile_ivar = ((tflativar[inorder] * tflat_fit^2) < 1.0e5)  * $
        (profile GT -0.1 AND profile LT 1.5 AND tflat_fit GT 20)


      profile_med = median(profile[xsort],15)
      outlier = (profile[xsort] - profile_med)^2*profile_ivar[xsort] LT 100
      totalivar = total(profile_ivar[xsort]*outlier) / total(outlier)
      slit_samp_use = totalivar GT 4000.0 ? slit_samp : slit_samp * 1.5
     
      bkpts0 = 0 
      profile_set = bspline_iterfit(slit_frac[xsort], profile[xsort], $
                                    invvar=profile_ivar[xsort]*outlier, $
                                    upper=2, lower=2, $
                                    /groupbadpix, maxrej=10, bkpt=bkpts0, $
                                    bkspace=slit_samp_use, yfit=profile_fit, $
                                    outmask=profile_mask, /silent) 

      if size(profile_set,/tname) EQ 'INT' then begin
          print, 'Could not fit 1d bspline'
          continue
      endif

      if keyword_set(chk) then plot, ywave, tflat_fit, ps=3, /xs

      ;; JXP: Above fit looks a little low? 
      ;; Let's try just a slope fit, with different bkpts...
      ;; All fits so far look like we could independently fit for the slope.

      profile_res = (profile[xsort] - profile_fit) / (y[xsort]+(y[xsort] EQ 0))
      profile_res_ivar = (profile_ivar[xsort]) * y[xsort]^2
      inner = where(abs(bkpts0) LT 0.8, n_inner)
      bkpts1 = bkpts0[where(abs(bkpts0) GE 0.8)]
      if n_inner GE 2 then bkpts1 = [bkpts1, $
                                     bkpts0[inner[2*lindgen(n_inner/2-1)+1]]]
      bkpts1 = bkpts1[sort(bkpts1)]

      profile_setslope = bspline_iterfit(slit_frac[xsort], profile_res, $
                                 invvar=profile_res_ivar*outlier, $
                                 bkpt=bkpts1, yfit=profile_fit_slope, $
                                 upper=2, lower=2, $
                                 /groupbadpix, maxrej=10, $
                                 outmask=profile_mask2d, /silent) 
      
      if size(profile_set2d,/tname) EQ 'INT' then begin
          print, 'Could not fit 2d bspline'
          continue
      endif

      profile_fit2d = profile_fit + profile_fit_slope*y[xsort]

      scatter1d = total(abs((profile[xsort] - profile_fit) *profile_mask) +$
                        (1-profile_mask))/ (total(profile_mask)+1)
      scatter2d = total(abs((profile[xsort] - profile_fit2d)*profile_mask2d)+$
                        (1-profile_mask2d))/ (total(profile_mask2d)+1)
      
      print, scatter1d, scatter2d, totalivar, slit_samp_use, n_elements(bkpts1)
     
          
      if (totalivar LT 1000.0) then begin
          if (totalivar LT 200.0) then begin
              print, 'mike_slitflat: Not enough signal to fit anything'
              print, 'mike_slitflat: Skipping for now'
          endif else begin
              print, 'mike_slitflat: Not enough signal to fit 2d'
              ordr_str[q].profile0  = bspline_valu(profile_x, profile_set) * $
                (profile_x LT max(slit_frac[xsort]*profile_mask) AND $
                 profile_x GT min(slit_frac[xsort]*profile_mask))
              prof_image[inorder[xsort]] = profile_fit * tflat_fit[xsort]
              residual[inorder[xsort]] = (profile[xsort] - profile_fit)* $
                sqrt(profile_ivar[xsort] * profile_mask) 
          endelse
      endif else begin
          
;
;    This is the linear change with row  (over nrow/2)
;
          ordr_str[q].profile0  = bspline_valu(profile_x, profile_set) * $
            (profile_x LT max(slit_frac[xsort]*profile_mask2d) AND $
             profile_x GT min(slit_frac[xsort]*profile_mask2d))
          ordr_str[q].profile1  = bspline_valu(profile_x, profile_setslope) * $
            (profile_x LT max(slit_frac[xsort]*profile_mask2d) AND $
             profile_x GT min(slit_frac[xsort]*profile_mask2d))
          
          prof_image[inorder[xsort]] = profile_fit2d * tflat_fit[xsort]
          residual[inorder[xsort]] = (profile[xsort] - profile_fit2d)* $
            sqrt(profile_ivar[xsort] * profile_mask2d) 
      endelse

      djs_iterstat, (profile[xsort] - profile_fit2d)* $
        sqrt(profile_ivar[xsort]), sigma=sig2d, sigrej=5.
      djs_iterstat, (profile[xsort] - profile_fit2d)* $
        sqrt(profile_ivar[xsort]*profile_mask2d),sigma=sig2d_clean, sigrej=5.
      djs_iterstat, (profile[xsort] - profile_fit)* $
        sqrt(profile_ivar[xsort]), sigma=sig, sigrej=5.
      djs_iterstat, (profile[xsort] - profile_fit)* $
        sqrt(profile_ivar[xsort]*profile_mask), sigma=sig_clean, sigrej=5.
    
      x = slit_frac[xsort] 
      ab = func_fit(x, profile_fit, 2, $
           invvar=profile_ivar*(abs(x) LT 0.7), function_name='poly')

      print, total([[outlier],[profile_mask],[profile_mask2d]],1), $
        [ab, sig,sig_clean, sig2d, sig2d_clean], format='(3(i7),6(f8.4))'
      
      temp_qa = slitflat_qa(slit_frac[xsort], profile[xsort], profile_mask, $
                     ordr_str[q])
      temp_qa.ab = ab
      temp_qa.npix = n_elements(profile_mask)
      temp_qa.nrej = total(profile_mask EQ 0)
      temp_qa.chi2 = sig_clean^2

      qa_str = struct_append(qa_str, temp_qa)

      if keyword_set(chk) then begin 
          plot, slit_frac, profile, ps=3, yr=[0.9, 1.1], color=clr.white, $
            background=clr.black, xr=[-1.3,1.3], /xs
          oplot, slit_frac[xsort], profile_fit2d, ps=3, color=clr.red
          oplot, slit_frac[xsort], profile_fit, ps=3, color=clr.blue
          oplot, profile_x, ordr_str[q].profile0, color=clr.green
          oplot, [-1.3,1.3], poly([-1.3,1.3],ab), color=clr.white, thick=3
          
          oplot, slit_frac, profile+1, ps=3
          oplot, slit_frac[xsort], profile_fit2d+1, ps=3, color=clr.red
          oplot, slit_frac[xsort], profile_fit+1, ps=3, color=clr.blue
          oplot, profile_x, ordr_str[q].profile0+1, color=clr.green
      endif

      ;; remove slope

      if keyword_set(detilt) then begin
        ss = poly(profile_x, ab)
        ordr_str[q].profile0 = ordr_str[q].profile0 / (ss + (ss EQ 0)) * $
                    (ss GT 0.5) 
        ordr_str[q].profile1 = ordr_str[q].profile1 / (ss + (ss EQ 0)) * $
                    (ss GT 0.5) 
      endif
  endfor

;
;    Fix any bad orders
;
  ordermask  = total(ordr_str.profile0,1) GT 0
  badorders = where(ordermask EQ 0, nbad)
  
  if nbad GT 0 then begin
      print, 'mike_slitflat: Fixing profiles with extrapolation: ', $
        strcompress(strjoin(string(badorders)))
      y0 = transpose(ordr_str.profile0)
      y1 = transpose(ordr_str.profile1)
      x = lindgen(nord) # replicate(1,251)
      ymask = (y0 NE 0)
      xy2traceset, x, y0, tset0, ncoeff=2, yfit=yfit0, invvar=ymask
      ordr_str[badorders].profile0 = transpose(yfit0[badorders,*])
      qa_str[badorders].mid = ordr_str[badorders].profile0
      qa_str[badorders].edgel = ordr_str[badorders].profile0
      qa_str[badorders].edger = ordr_str[badorders].profile0
   
  endif

  return, ordr_str
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

pro mike_slitflat, mike, setup, side, chk=chk, $
            tflat_fil=tflat_fil, profile_fil=iprofile_fil, $
            jacob_fil=ijacob_fil, CLOBBER=clobber, DETILT=detilt

;
  if  N_params() LT 2  then begin 
      print,'Syntax - ' + $
        'mike_slitflat, mike, setup, [side], /clobber, /chk, /DETILT' 
      print, '    JACOB_FIL=, TFLAT_FIL=, PROFILE_FIL=  [v1.1]'
      return
  endif

   if not keyword_set( SIDE ) then side = [1L,2L]
   if not keyword_set( chk  ) then chk=0

; Setup
   if setup GE 10 then c_s = strtrim(setup,2) else c_s = '0'+strtrim(setup,2) 

;  Loop on side
   for ii=0L,n_elements(side)-1 do begin
       qq = side[ii]
       ;; SIDE
       if qq EQ 1 then begin
           print, 'mike_slitflat: Profiling BLUE trace flat'
           nm = 'B'
       endif else begin
           print, 'mike_slitflat: Profiling RED trace flat'
           nm = 'R' 
       endelse

       ;; Order structure
       ordr_str   = mike_getfil('ordr_str', setup, SIDE=qq)

       ;; Process an arc if necessary!
       if ordr_str[0].arc_m[0] EQ 0. then begin
           ;; Find the arc
           print, 'mike_slitflat: Processing an Arc.  Wish me luck!'
           arc = where(mike.type EQ 'ARC' AND mike.flg_anly NE 0 AND $
                       mike.setup EQ setup AND mike.side EQ qq, narc)
           if narc EQ 0 then begin
              print, "mike_slitflat: no arcs to be found"
              continue
           endif

           
           for iarc=0,n_elements(arc) do begin
             fil_nm = mike_getfil('arc_trc', setup, $
                 subfil=mike[arc[iarc]].img_root, /name, chkf=chkf)
             if chkf EQ 1 then begin
               arc_fil = mike[arc[iarc]].rootpth + mike[arc[iarc]].img_root
               break
             endif
           endfor

           ;; Process
           resolve_routine, 'mike_allarc', /no_recompile, /either

           ;; likely we only need to refit arc trace...

           if NOT keyword_set(clobber) then $
             rslt = mike_fittrcarc_work( arc_fil, setup, qq, CLOBBER=1) $
           else  $
             rslt = mike_allarc_sngl(arc_fil, setup, qq, CLOBBER=1)

           if rslt EQ '' then stop
           ;; Reread order structure and check
           ordr_str   = mike_getfil('ordr_str', setup, SIDE=qq)
           if ordr_str[0].arc_m[0] EQ 0. then stop
       endif
           
       ;; TFLAT
       tflat     = mike_getfil('tflat_fil',setup, SIDE=qq)
       tflativar = mike_getfil('tflat_fil',setup, SIDE=qq, INDX=1)


       print, 'mike_slitflat: NOT Applying MFlat..., already done in mktflat'
       ;;
       ;; MFLAT
       ;; mflat = mike_getfil('mflat_fil', setup, SIDE=qq)
       ;; tflat = tflat / (mflat + (mflat LE 0.5))
       ;; tflativar = tflativar * mflat^2 * (mflat GT 0.5) 

       infinite = where(finite(tflat) EQ 0)
       if infinite[0] NE -1 then begin
         tflat[infinite] = 0
         tflativar[infinite] = 0
       endif


       if NOT keyword_set(iprofile_fil) then $
         profile_fil = mike_getfil('ordr_str', setup, SIDE=qq, /name) $
       else profile_fil = iprofile_fil

       if NOT keyword_set(iproimg_fil) then $
         proimg_fil = 'Flats/Profile_'+nm+'_'+c_s+'.fits' $ 
       else proimg_fil = iproimg_fil

       if NOT keyword_set(ijacobian_fil) then $
         jacobian_fil = 'Flats/Jacobian_'+nm+'_'+c_s+'.fits' $
         else jacobian_fil = ijacobian_fil


       nxbkpt = qq EQ 1 ? 8 : 32
       nybkpt = qq EQ 1 ? 5 : 8

       profile_struct = mike_slitflat_work(tflat, tflativar, ordr_str, $
                       jacobian=jacobian, chk=chk, residual=residual, $
                       nxbkpt=nxbkpt, nybkpt=nybkpt, qa_str=qa_str, $
                       prof_image=profile_img, detilt=keyword_set(detilt))

       ;; Write to fits
       print, 'mike_slitflat: Writing order w/Profile structure ', profile_fil
       mwrfits, profile_struct, profile_fil, /create
       
       ;; Write to fits
       print, 'mike_slitflat: Writing Profile Image structure ', proimg_fil
       mwrfits, profile_img, proimg_fil, /create
       
       if keyword_set(jacobian) then begin
           print, 'mike_slitflat: Writing Jacobian Image', jacobian_fil
           mwrfits, jacobian, jacobian_fil, /create
           mwrfits, residual, jacobian_fil
       endif

       qafil = mike_getfil('qa_slitflat', setup, SIDE=qq)
       qa_profile, qa_str, qafil

   endfor

   ;; All done
   print, 'mike_slitflat: All done'
   return

end


